(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontmodule.ml,v 1.3 2002/02/07 14:58:28 malekith Exp $
 *
 * Module implementation: Gontmodule
 * Description: We implement module interface here. As Gont modules are
 * very simple, so *this* module is :)
 *)

open Gontutil

type mv_record = {mv_name: id;
                  mv_module: id;
                  mv_type: ty;
                  mv_props: symbol_property list;
                  mv_union_type: typedef option;
                  mv_field_id: int;
                  mutable mv_symbol: symbol option}

type mt_record = {mt_name: id;
                  mt_tydef: typedef}
                  
type mod_record = {mod_name: id;
                   mod_vals: mv_record list;
                   mod_types: mt_record list}

(* We use it as magic. It is stored in all .gio files, if file contain
 * bad ident string -- it is rejected. This is probably not enough 
 * (we should also put version of gontutil.mli here, but... nothing's
 * perfect :^) *)
let ver = "Gont Compiled Interface, $Revision: 1.3 $"

let strip_module mod_name sym_name =
  let len = (String.length mod_name) + 2 in
  let slen = String.length sym_name in
  if slen > len && 
     (String.sub sym_name 0 len) = (mod_name ^ "::") then
    String.sub sym_name len (slen - len)
  else
    sym_name 

let save_module out_ch mod_name symbols types =
  let make_mv sym =
    assert (sym#get_fnc_level == 0);
    assert (sym#get_scope_level == 0);
    {mv_name = strip_module mod_name sym#get_name;
     mv_module = mod_name;
     mv_type = sym#get_type;
     mv_props = sym#properties;
     mv_union_type = 
       if sym#has_property IsUnionField then Some sym#union_type
       else None;
     mv_field_id =
       if sym#has_property IsUnionField then sym#field_id else -1;
     mv_symbol = None} in
  let make_mt ({td_name = n} as tydef) =
    {mt_name = strip_module mod_name n; 
     mt_tydef = {tydef with td_module = mod_name}} in
  let r = {mod_name = mod_name; 
           mod_vals = List.map make_mv symbols;
           mod_types = List.map make_mt types} in
  Marshal.to_channel out_ch ver [];
  Marshal.to_channel out_ch r []
;;

let modules : (string, mod_record) Hashtbl.t = Hashtbl.create 16

let reset () = Hashtbl.clear modules

exception Cannot_open
    
let open_module m =
  let load m =
    let rec open_any names =
      match names with
          x::xs -> 
            (try
              open_in x
             with 
               Sys_error _ -> open_any xs)
        | [] -> raise Cannot_open in
    let in_ch = 
      try
        open_any [((String.uncapitalize m) ^ ".gio")]
      with Cannot_open -> 
        raise (Error (xf "cannot find module interface for %s" m)) in
    let id = (Marshal.from_channel in_ch : string) in
      if id <> ver then
        (close_in in_ch; 
         raise (Error (xf "module interface file for %s was compiled \
                           with diffrent version, please recompile \
                           the interface" m)))
      else
        let mr = (Marshal.from_channel in_ch : mod_record) in
          if mr.mod_name <> m then
            (close_in in_ch;
             raise (Error (xf "interface file contains module %s, \
                               instead of expected %s" mr.mod_name m)))
           else
             Hashtbl.add modules mr.mod_name mr;
             close_in in_ch in
  try
    ignore(Hashtbl.find modules m)
  with Not_found -> load m
;;

let find_module m =
  open_module m;
  Hashtbl.find modules m

exception Symbol_found of symbol
exception Tydef_found of typedef

let get_symbol m =
   match m with
   | {mv_symbol = Some sym} ->
       sym
   | _ ->
       let sym = new_symbol (0, 0, m.mv_type, 
                                   m.mv_module ^ "::" ^ m.mv_name) in
        (match m.mv_union_type with
         | Some t -> sym#make_union_field (m.mv_field_id, t)
         | None -> ());
       List.iter (fun p -> sym#set_property p) m.mv_props;
       m.mv_symbol <- Some sym;
       sym
;;

let import_symbol mod_name sym_name =
  try
    List.iter (fun s -> 
       match s with
       | {mv_name = n} when (n = sym_name) ->
            raise (Symbol_found (get_symbol s))
       | _ -> ()) (find_module mod_name).mod_vals;
    raise Not_found
  with Symbol_found s -> s
  
let import_type_symbol mod_name sym_name =
  try
    List.iter (fun s -> 
       match s with
          {mt_name = n; mt_tydef = t} when (n = sym_name) -> 
            raise (Tydef_found t)
        | _ -> ()) (find_module mod_name).mod_types;
    raise Not_found
  with Tydef_found t -> t
    
let globals mod_name =
  let m = find_module mod_name in
  ((List.map get_symbol m.mod_vals),
   (List.map (fun {mt_tydef = t} -> t) m.mod_types))
;;
