(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontutil.ml,v 1.12 2002/02/12 11:52:41 malekith Exp $
 *
 * Module interface: Gontutil
 * Description: Various utilties and datatypes defintion for Gont compiler.
 *)

(* Identifier. *)
type id = string

type location = {mutable loc_file: string; 
                 mutable loc_char: int;
                 mutable loc_line: int}

(* Types used by Gont compiler. *)
type ty =
    Int                         (* integer of the same size as pointer *)
  | Void                        (* unit type *)
  | Bool                        (* boolean, same size as Int *)
  | Float                       (* C's double *)
  | String
  | Tuple of ty list
  | Func of ft_record
  | TyVar of id
  | Struct of (ty * id) list
  | OptStruct of (ty * id) list
  | Union of (ty * id) list
  | AbstractType
  | NamedType of nt_record      (* <int>List::t, some_struct *)
and typedef = 
    {mutable td_tyvars: id list; 
     mutable td_type: ty;
     mutable td_module: id;         (* "" for current module, until Gontbind *)
     td_name: id;
     td_loc: location}
and nt_record = 
    {nt_parms: ty list;
     nt_module: id;
     nt_name: id;
     nt_loc: location;
     mutable nt_tydef: typedef option; (* Filled by Gontbind. *)
   }
and ft_record = {ft_ret_type: ty;
                 ft_parms: ty list;
                 mutable ft_tyvars: id list}
;;

type symbol_property =
    IsInClosure
  | IsUnionField
  | IsTopPattern
  | IsNonLvalue
  | IsPrototyped
  | IsNonLocal
  | IsInitialized
  | IsInPattern
  | IsGlobalVar
  
class type symbol = 
  object
    method get_name : id
    
    method get_type : ty
    method set_type : ty -> unit
    
    method get_fnc_level : int
    method get_scope_level : int
    
    method has_property : symbol_property -> bool
    method set_property : symbol_property -> unit
    method unset_property : symbol_property -> unit
    method properties : symbol_property list
    
    method make_union_field : int * typedef -> unit
    method union_type : typedef
    method field_id : int
  end

type op2 =
    IntPlus
  | IntMinus
  | IntMult
  | IntDiv
  | IntMod
  | FloatPlus
  | FloatMinus
  | FloatMult
  | FloatDiv
  | Assign
  | Eq | Ne | Lt | Gt | Ge | Le
  
type op1 =
    IntNegate
  | FloatNegate
  
type parm_record = {parm_name: id;
                    parm_type: ty;
                    mutable parm_symbol: symbol option}

type raw_exp =
    SimpleBinOp of op2 * exp * exp
  | SimpleUnOp of op1 * exp
  | Ref of ref_record   (* reference to variable *)
  | Call of exp * exp list
  | IntConst of string
  | BoolConst of bool
  | FloatConst of string
  | VoidConst
  | StringConst of string
  | TupleExp of exp list
  | UnionExp of union_record
  | Lambda of lambda_record
  | CrossModuleRef of xmr_record
and exp = {mutable raw_exp: raw_exp; 
           mutable exp_type: ty option; 
           exp_loc: location}
and ref_record = {ref_name: id;
                  mutable ref_level: int;
                  mutable ref_symbol: symbol option}
and xmr_record = {xmr_module: id;
                  xmr_name: id;
                  mutable xmr_symbol: symbol option}
and union_record = {ur_exp: exp;
                    ur_module: id;
                    ur_name: id;
                    mutable ur_sym: symbol option}
and raw_stmt =
    Skip
  | Exp of exp
  | Return of exp
  | ReturnVoid
  | VarDef of var_record
  | Block of stmt list
  | Pair of stmt * stmt         (* doesn't introduce new scope *)
  | Switch of exp * (case_clouse list)
  | Let of ((pattern * exp) list) * stmt
  | If of exp * stmt * stmt
  | Loop of stmt
  | BreakIfNot of exp
  | Open of id
and stmt = {raw_stmt: raw_stmt;
            stmt_loc: location}
and var_record = {var_name: id;
                  var_type: ty;
                  mutable var_symbol: symbol option}
and lambda_record = {lambda_ret_type: ty;
                     lambda_parms: parm_record list;
                     lambda_body: stmt;
                     lambda_tyvars: id list}
and case_clouse = {cc_pattern: pattern;
                   cc_stmt: stmt}
and pattern = {mutable raw_pattern: raw_pattern;
               mutable pat_loc: location}
and raw_pattern = 
    VoidPattern
  | VarPattern of var_pattern_record
  | ExprPattern of exp
  | TuplePattern of tuple_pattern_record
  | UnionPattern of union_pattern_record
and tuple_pattern_record = {tpr_patterns: pattern list}
and union_pattern_record = {upr_module: id;
                            upr_name: id;
                            mutable upr_sym: symbol option;
                            upr_pattern: pattern}
and var_pattern_record = {vp_name: id;
                  mutable vp_symbol: symbol option}


type fn_decl = {fn_ret_type: ty; 
                fn_name: id; 
                fn_parms: parm_record list;
                fn_body: stmt;
                fn_loc: location;
                fn_tyvars: id list;
                mutable fn_symbol: symbol option;
                fn_export_as: string}

type extern_c_decl = {xc_type: ty; 
                      xc_name: id; 
                      xc_loc: location;
                      mutable xc_symbol: symbol option}

type global_decl = {global_type: ty;
                    global_loc: location;
                    global_name: id;
                    mutable global_symbol: symbol option}

type fn_proto_decl = {fp_type: ty;
                      fp_name: id;
                      fp_loc: location;
                      mutable fp_symbol: symbol option}
                
type decl =
    FuncDecl of fn_decl
  | GlobalDecl of global_decl
  | ExternC of extern_c_decl
  | FuncProto of fn_proto_decl
  | TopLevelOpen of id
  | TypeDef of typedef

exception Syntax_error of location * string
exception Error of string
exception Eof
;;

(* Utils. *)
let warnings_count = ref 0

let xf = Printf.sprintf

let err_msg (loc, msg) =
  prerr_string (xf "%s:%d:%d: %s" 
                   loc.loc_file loc.loc_line (loc.loc_char+1)
                   msg);
  prerr_newline ()

let warn (loc, msg) = 
  incr warnings_count;
  err_msg (loc, ("warning: " ^ msg))

let err (loc, msg) = raise (Syntax_error(loc, msg))

let debug s = print_string s; print_newline ()

let rec list_apply fnc (c, xxs) =
  match xxs with
      x::xs -> 
        let (c', x') = fnc (c, x) in
        let (c'', xs') = list_apply fnc (c', xs) in
        (c'', x'::xs')
    | [] -> (c, [])

let sequences : (string, int ref) Hashtbl.t = Hashtbl.create 16

let reset () = Hashtbl.clear sequences

let seq_name s =
  let r = 
    (try
       Hashtbl.find sequences s 
     with Not_found -> 
       let n = ref 0 in
       Hashtbl.add sequences s n; n) in
  incr r; s ^ (string_of_int !r)

let rec type_walk f t =
  let fs = type_walk f in
  let fl = List.map fs in
  let t' = f t in
  match t' with
      Int | Void | Bool | Float | String | TyVar _ | AbstractType -> t'
    | Tuple l -> Tuple (fl l)
    | Func {ft_ret_type = r; ft_parms = p; ft_tyvars = t} ->
       Func {ft_ret_type = fs r; ft_parms = fl p; ft_tyvars = t}
    | Struct l -> Struct (List.map (fun (t, n) -> (fs t, n)) l)
    | OptStruct l -> OptStruct (List.map (fun (t, n) -> (fs t, n)) l)
    | Union l -> Union (List.map (fun (t, n) -> (fs t, n)) l)
    | NamedType ({nt_parms = p} as r) -> 
       NamedType {r with nt_parms = fl p}

let simple_fv t =
  let res : id list ref = ref [] in
  let f t =
    match t with
        TyVar(a) ->
          if not (List.mem a !res) then
            res := a :: !res;
          t
      | NamedType _ | Int | Void | Bool | Float | String | AbstractType
      | Tuple _ | Func _ | Struct _ | OptStruct _ | Union _ -> t in
  ignore(type_walk f t);
  !res

let make_fn_type (r, a) =
  let a' = List.map (fun {parm_type = t} -> t) a in
  let t = {ft_ret_type = r; ft_parms = a'; ft_tyvars = []} in
  let fv = simple_fv (Func t) in
  Func {t with ft_tyvars = fv}
;;

class symbol_impl = fun (scope, fnc, t, n) ->
  object
    val mutable sym_type : ty = t
    val sym_name : id = n
    val sym_scope_level : int = scope
    val sym_fnc_level : int = fnc
    method get_name = sym_name
    method get_type = sym_type
    method set_type t = sym_type <- t
    
    method get_fnc_level = sym_fnc_level
    method get_scope_level = sym_scope_level

    val mutable props : symbol_property list = []

    val mutable sym_union_field = -1
    val mutable sym_union_type : typedef option = None
    method make_union_field (i, t) =
       assert (not (List.memq IsUnionField props));
       props <- IsUnionField :: props;
       sym_union_field <- i;
       sym_union_type <- Some t
    method field_id = 
       assert (sym_union_field != -1); 
       sym_union_field
    method union_type =
      match sym_union_type with
      | None -> assert false
      | Some x -> x
    
    method set_property p =
       if not (List.memq p props) then
         props <- p :: props
    method has_property p =
       List.memq p props
    method unset_property p =
      props <- List.filter (fun x -> x != p) props
    method properties = props
  end

let new_symbol x = 
  let res : symbol = new symbol_impl x in
  res
;;
